/*****************************************************************************
** FILE IDENTIFICATION
**
**   Name:          ctsim.cpp
**   Purpose:       Top-level routines of CTSim program
**   Programmer:    Kevin Rosenberg
**   Date Started:  July 2000
**
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: ctsim.cpp,v 1.21 2000/12/27 03:16:02 kevin Exp $
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/


// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#if !wxUSE_DOC_VIEW_ARCHITECTURE
#error You must set wxUSE_DOC_VIEW_ARCHITECTURE to 1 in setup.h!
#endif

#include "ct.h"
#include "ctsim.h"
#include "docs.h"
#include "views.h"
#include "dialogs.h"

#if defined(HAVE_CONFIG_H)
#include "config.h"
#endif

#if defined(HAVE_GETOPT_H) || defined(HAVE_GETOPT_LONG)
#ifdef MSVC
#define __STDC__ 1
#endif
#include "getopt.h"
#ifdef MSVC
#undef __STDC__
#endif
#endif

static const char* rcsindent = "$Id: ctsim.cpp,v 1.21 2000/12/27 03:16:02 kevin Exp $";

class CTSimApp* theApp = NULL;

struct option CTSimApp::ctsimOptions[] = 
{
    {"help", 0, 0, O_HELP},
    {"version", 0, 0, O_VERSION},
    {0, 0, 0, 0}
};

IMPLEMENT_APP(CTSimApp)

CTSimApp::CTSimApp()
  : m_docManager(NULL), m_pFrame(NULL)
{
    theApp = this;
}

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

bool
CTSimApp::OnInit()
{
#ifdef HAVE_SETPRIORITY
  setpriority (PRIO_PROCESS, 0, 15);  // set to low scheduling priority
#endif

    // process options
    while (1) {
      int c = getopt_long (argc, argv, "", ctsimOptions, NULL);
      if (c == -1)
	break;

      switch (c) {
      case O_VERSION:
		  std::cout << rcsindent << std::endl;
#ifdef CTSIMVERSION
		  std::cout << "Version: CTSIMVERSION" << std::endl;
#elif defined(VERSION)
		  std::cout << "Version: VERSION" << std::endl;
#endif
	  exit(0);
      case O_HELP:
      case '?':
	usage (argv[0]);
	exit (0);
      default:
	usage (argv[0]);
	exit (1);
      }
    }

    m_docManager = new wxDocManager;
    
    new wxDocTemplate (m_docManager, "ImageFile", "*.if", "", "if", "ImageFile doc", "ImageFile View", CLASSINFO(ImageFileDocument), CLASSINFO(ImageFileView));

    new wxDocTemplate (m_docManager, "ProjectionFile", "*.pj", "", "pj", "ProjectionFile doc", "ProjectionFile View", CLASSINFO(ProjectionFileDocument), CLASSINFO(ProjectionFileView));

    new wxDocTemplate (m_docManager, "PhantomFile", "*.phm", "", "phm", "Phantom doc", "Phantom View", CLASSINFO(PhantomDocument), CLASSINFO(PhantomView));

    new wxDocTemplate (m_docManager, "PlotFile", "*.plt", "", "plt", "Plot doc", "Plot View", CLASSINFO(PlotFileDocument), CLASSINFO(PlotFileView));

    //// Create the main frame window
    m_pFrame = new MainFrame(m_docManager, (wxFrame *) NULL, -1, "CTSim", wxPoint(0, 0), wxSize(500, 400), wxDEFAULT_FRAME_STYLE);
    
    SetTopWindow (m_pFrame);
    m_pFrame->Centre(wxBOTH);

    m_pFrame->Show(true);

    for (int i = optind + 1; i <= argc; i++) {
      wxString filename = argv [i - 1];
      m_docManager->CreateDocument (filename, wxDOC_SILENT);
    }

    return true;
}

void
CTSimApp::usage(const char* program)
{
	std::cout << "usage: " << fileBasename(program) << " [files-to-open...] [OPTIONS]\n";
	std::cout << "Computed Tomography Simulator (Graphical Shell)\n";
	std::cout << "\n";
	std::cout << "  --version Display version\n";
	std::cout << "  --help    Display this help message\n";
}

int
CTSimApp::OnExit()
{
    delete m_docManager;
#ifdef HAVE_DMALLOC
    dmalloc_shutdown();
#endif
    return 0;
}

wxString
CTSimApp::getUntitledFilename()
{
  static int untitledNumber = 1;

  wxString filename ("Untitled");
  filename << untitledNumber++;

  return (filename);
}


// Top-level window for CTSim

IMPLEMENT_CLASS(MainFrame, wxDocParentFrame)

BEGIN_EVENT_TABLE(MainFrame, wxDocParentFrame)
  EVT_MENU(MAINMENU_HELP_ABOUT, MainFrame::OnAbout)
  EVT_MENU(MAINMENU_HELP_CONTENTS, MainFrame::OnHelpContents)
  EVT_MENU(MAINMENU_FILE_CREATE_PHANTOM, MainFrame::OnCreatePhantom)
  EVT_MENU(MAINMENU_FILE_EXIT, MainFrame::OnExit)
  EVT_MENU(MAINMENU_WINDOW_BASE, MainFrame::OnWindowMenu0)
  EVT_MENU(MAINMENU_WINDOW_BASE+1, MainFrame::OnWindowMenu1)
  EVT_MENU(MAINMENU_WINDOW_BASE+2, MainFrame::OnWindowMenu2)
  EVT_MENU(MAINMENU_WINDOW_BASE+3, MainFrame::OnWindowMenu3)
  EVT_MENU(MAINMENU_WINDOW_BASE+4, MainFrame::OnWindowMenu4)
  EVT_MENU(MAINMENU_WINDOW_BASE+5, MainFrame::OnWindowMenu5)
  EVT_MENU(MAINMENU_WINDOW_BASE+6, MainFrame::OnWindowMenu6)
  EVT_MENU(MAINMENU_WINDOW_BASE+7, MainFrame::OnWindowMenu7)
  EVT_MENU(MAINMENU_WINDOW_BASE+8, MainFrame::OnWindowMenu8)
  EVT_MENU(MAINMENU_WINDOW_BASE+9, MainFrame::OnWindowMenu9)
  EVT_MENU(MAINMENU_WINDOW_BASE+10, MainFrame::OnWindowMenu10)
  EVT_MENU(MAINMENU_WINDOW_BASE+11, MainFrame::OnWindowMenu11)
  EVT_MENU(MAINMENU_WINDOW_BASE+12, MainFrame::OnWindowMenu12)
  EVT_MENU(MAINMENU_WINDOW_BASE+13, MainFrame::OnWindowMenu13)
  EVT_MENU(MAINMENU_WINDOW_BASE+14, MainFrame::OnWindowMenu14)
  EVT_MENU(MAINMENU_WINDOW_BASE+15, MainFrame::OnWindowMenu15)
  EVT_MENU(MAINMENU_WINDOW_BASE+16, MainFrame::OnWindowMenu16)
  EVT_MENU(MAINMENU_WINDOW_BASE+17, MainFrame::OnWindowMenu17)
  EVT_MENU(MAINMENU_WINDOW_BASE+18, MainFrame::OnWindowMenu18)
  EVT_MENU(MAINMENU_WINDOW_BASE+19, MainFrame::OnWindowMenu19)
  EVT_UPDATE_UI_RANGE(MAINMENU_WINDOW_BASE, MAINMENU_WINDOW_BASE+20, MainFrame::OnUpdateUI)
END_EVENT_TABLE()



MainFrame::MainFrame(wxDocManager *manager, wxFrame *frame, wxWindowID id, const wxString& title, const wxPoint& pos, const wxSize& size, const long type)
  : wxDocParentFrame(manager, frame, id, title, pos, size, type), m_pLog(NULL)
{
    m_pLog = new wxTextCtrl (this, -1, "Log Window\n", wxPoint(0, 250), wxSize(100,50), wxTE_MULTILINE | wxTE_READONLY);
    wxLog::SetActiveTarget(new wxLogTextCtrl(m_pLog));
    CreateStatusBar();
    SetStatusText ("Welcome to CTSim");

    //// Make a menubar
    wxMenu *file_menu = new wxMenu;
    
    file_menu->Append(MAINMENU_FILE_CREATE_PHANTOM, "Cr&eate Phantom...");
    file_menu->Append(wxID_OPEN, "&Open...");
    
    file_menu->AppendSeparator();
    file_menu->Append(MAINMENU_FILE_EXIT, "E&xit");
    
    //  history of files visited
    m_docManager->FileHistoryUseMenu(file_menu);
    
    m_pWindowMenu = new wxMenu;
    m_pWindowMenu->UpdateUI (this);

    wxMenu* help_menu = new wxMenu;
    help_menu->Append(MAINMENU_HELP_CONTENTS, "&Contents");
    help_menu->AppendSeparator();
    help_menu->Append(MAINMENU_HELP_ABOUT, "&About");
    
    wxMenuBar* menu_bar = new wxMenuBar;
    
    menu_bar->Append(file_menu, "&File");
    menu_bar->Append(m_pWindowMenu, "&Window");
    menu_bar->Append(help_menu, "&Help");
    
    SetMenuBar(menu_bar);

    for (int i = 0; i < MAX_WINDOW_MENUITEMS; i++) {
      m_apWindowMenuItems[i] = new wxMenuItem (m_pWindowMenu, MAINMENU_WINDOW_BASE+i, wxString("<Empty>"));
      m_pWindowMenu->Append (m_apWindowMenuItems[i]);
      m_pWindowMenu->Enable (MAINMENU_WINDOW_BASE+i, false);
    }
}

void 
MainFrame::OnAbout(wxCommandEvent& WXUNUSED(event) )
{
	wxString msg = "CTSim\nThe Open Source Computed Tomography Simulator\n";
#ifdef CTSIMVERSION
	msg += "Version ";
	msg += CTSIMVERSION;
	msg += "\n\n";
#elif defined(VERSION)
	msg << "Version: " <<  VERSION << "\n\n";
#endif
	msg += "Author: Kevin Rosenberg <kevin@rosenberg.net>\nUsage: ctsim [files-to-open..] [--help]";

    wxMessageBox(msg, "About CTSim", wxOK | wxICON_INFORMATION, this);
}

void 
MainFrame::OnCreatePhantom(wxCommandEvent& WXUNUSED(event))
{
    DialogGetPhantom dialogPhantom (this, Phantom::PHM_HERMAN);
    int dialogReturn = dialogPhantom.ShowModal();
    if (dialogReturn == wxID_OK) {
      wxString selection (dialogPhantom.getPhantom());
      *theApp->getLog() << "Selected phantom " << selection.c_str() << "\n";
      wxString filename = selection + ".phm";
      theApp->getDocManager()->CreateDocument(filename, wxDOC_SILENT);
    }
    
}

void
CTSimApp::getCompatibleImages (const ImageFileDocument* pIFDoc, std::vector<ImageFileDocument*>& vecIF)
{
  const ImageFile& rIF = pIFDoc->getImageFile();
  int nx = rIF.nx();
  int ny = rIF.ny();
  wxList& rListDocs = m_docManager->GetDocuments();
  for (wxNode* pNode = rListDocs.GetFirst(); pNode != NULL; pNode = pNode->GetNext()) {
    wxDocument* pDoc = reinterpret_cast<wxDocument*>(pNode->GetData());
    ImageFileDocument* pIFCompareDoc = dynamic_cast<ImageFileDocument*>(pDoc);
    if (pIFCompareDoc && (pIFDoc != pIFCompareDoc)) {
      const ImageFile& rCompareIF = pIFCompareDoc->getImageFile();
      if (rCompareIF.nx() == nx && rCompareIF.ny() == ny)
        vecIF.push_back (pIFCompareDoc);
    }
  }
}

void 
MainFrame::OnHelpContents(wxCommandEvent& WXUNUSED(event) )
{
    wxMessageBox("No help available, refer to man pages of command-line tools");
}

void 
MainFrame::OnExit (wxCommandEvent& WXUNUSED(event) )
{
    Close(true);
}

void
MainFrame::OnUpdateUI (wxUpdateUIEvent& rEvent)
{
  int iPos = 0;
  wxList& rListDocs = m_docManager->GetDocuments();
  wxNode* pNode = rListDocs.GetFirst();
  while (iPos < MAX_WINDOW_MENUITEMS && pNode != NULL) {
    wxDocument* pDoc = static_cast<wxDocument*>(pNode->GetData());
    wxString strFilename = pDoc->GetFilename();
    static_cast<wxMenuItemBase*>(m_apWindowMenuItems[iPos])->SetName (strFilename);
    m_apWindowMenuData[iPos] = pDoc;
    m_pWindowMenu->Enable (MAINMENU_WINDOW_BASE+iPos, true);
    iPos++;
    pNode = pNode->GetNext();
  }
  for (int i = iPos; i < MAX_WINDOW_MENUITEMS; i++) {
    m_pWindowMenu->Enable (MAINMENU_WINDOW_BASE+i, false);
    static_cast<wxMenuItemBase*>(m_apWindowMenuItems[i])->SetName (wxString("<Empty>"));
    m_apWindowMenuData[i] = NULL;
  }
    
}

void 
MainFrame::DoWindowMenu (int iMenuPosition, wxCommandEvent& event)
{
  if (wxDocument* pDoc = m_apWindowMenuData [iMenuPosition]) {
    wxString strFilename = pDoc->GetFilename();
    const wxView* pView = pDoc->GetFirstView();
    if (pView) {
      wxFrame* pFrame = pView->GetFrame();
      pFrame->SetFocus();
      pFrame->Raise();
    }
  }
}

void MainFrame::OnWindowMenu0 (wxCommandEvent& event)
{ DoWindowMenu (0, event); }

void MainFrame::OnWindowMenu1 (wxCommandEvent& event)
{ DoWindowMenu (1, event); }

void MainFrame::OnWindowMenu2 (wxCommandEvent& event)
{ DoWindowMenu (2, event); }

void MainFrame::OnWindowMenu3 (wxCommandEvent& event)
{ DoWindowMenu (3, event); }

void MainFrame::OnWindowMenu4 (wxCommandEvent& event)
{ DoWindowMenu (4, event); }

void MainFrame::OnWindowMenu5 (wxCommandEvent& event)
{ DoWindowMenu (5, event); }

void MainFrame::OnWindowMenu6 (wxCommandEvent& event)
{ DoWindowMenu (6, event); }

void MainFrame::OnWindowMenu7 (wxCommandEvent& event)
{ DoWindowMenu (7, event); }

void MainFrame::OnWindowMenu8 (wxCommandEvent& event)
{ DoWindowMenu (8, event); }

void MainFrame::OnWindowMenu9 (wxCommandEvent& event)
{ DoWindowMenu (9, event); }

void MainFrame::OnWindowMenu10 (wxCommandEvent& event)
{ DoWindowMenu (10, event); }

void MainFrame::OnWindowMenu11 (wxCommandEvent& event)
{ DoWindowMenu (11, event); }

void MainFrame::OnWindowMenu12 (wxCommandEvent& event)
{ DoWindowMenu (12, event); }

void MainFrame::OnWindowMenu13 (wxCommandEvent& event)
{ DoWindowMenu (13, event); }

void MainFrame::OnWindowMenu14 (wxCommandEvent& event)
{ DoWindowMenu (14, event); }

void MainFrame::OnWindowMenu15 (wxCommandEvent& event)
{ DoWindowMenu (15, event); }

void MainFrame::OnWindowMenu16 (wxCommandEvent& event)
{ DoWindowMenu (16, event); }

void MainFrame::OnWindowMenu17 (wxCommandEvent& event)
{ DoWindowMenu (17, event); }

void MainFrame::OnWindowMenu18 (wxCommandEvent& event)
{ DoWindowMenu (18, event); }

void MainFrame::OnWindowMenu19 (wxCommandEvent& event)
{ DoWindowMenu (19, event); }


