/*****************************************************************************
** FILE IDENTIFICATION
**
**   Name:          doc.cpp
**   Purpose:       Document routines for CTSim program
**   Programmer:    Kevin Rosenberg
**   Date Started:  July 2000
**
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: docs.cpp,v 1.8 2000/12/27 03:16:02 kevin Exp $
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

#ifdef __GNUG__
// #pragma implementation
#endif

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif
#include "wx/txtstrm.h"
#include "wx/file.h"

#if !wxUSE_DOC_VIEW_ARCHITECTURE
#error You must set wxUSE_DOC_VIEW_ARCHITECTURE to 1 in setup.h!
#endif

#include "ct.h"
#include "ctsim.h"
#include "docs.h"
#include "views.h"

// ImageFileDocument

IMPLEMENT_DYNAMIC_CLASS(ImageFileDocument, wxDocument)

bool ImageFileDocument::OnSaveDocument(const wxString& filename)
{
  if (! m_imageFile.fileWrite (filename)) {
    *theApp->getLog() << "Unable to write image file " << filename << "\n";
    return false;
  }
  *theApp->getLog() << "Wrote image file " << filename << "\n";
  Modify(false);
  return true;
}

bool ImageFileDocument::OnOpenDocument(const wxString& filename)
{
  if (filename == "untitled.if") {
    wxString untitledFilename = theApp->getUntitledFilename();
    SetFilename (untitledFilename, true);
  } else {
    if (! m_imageFile.fileRead (filename.c_str())) {
      *theApp->getLog() << "Unable to read image file " << filename << "\n";
      return false;
    }
    *theApp->getLog() << "Read image file " << filename << "\n";
    SetFilename(filename, true);
  }
  Modify(false);
  UpdateAllViews();
  ImageFileView* ifView = dynamic_cast<ImageFileView*>(GetFirstView());
  if (ifView)
	  ifView->OnUpdate(ifView, NULL);

  return true;
}

bool ImageFileDocument::IsModified(void) const
{
  return wxDocument::IsModified();
}

void ImageFileDocument::Modify(bool mod)
{
    wxDocument::Modify(mod);
}

bool ImageFileDocument::OnCloseDocument ()
{
  bool bReturn = wxDocument::OnCloseDocument();
  return bReturn;
}

// ProjectionFileDocument

IMPLEMENT_DYNAMIC_CLASS(ProjectionFileDocument, wxDocument)

bool ProjectionFileDocument::OnSaveDocument(const wxString& filename)
{
  if (! m_projectionFile.write (filename.c_str())) {
    *theApp->getLog() << "Unable to write projection file " << filename << "\n";
    return false;
  }
  *theApp->getLog() << "Wrote projection file " << filename << "\n";
  Modify(false);
  return true;
}

bool ProjectionFileDocument::OnOpenDocument(const wxString& filename)
{
  if (filename == "untitled.pj") {
    wxString untitledFilename = theApp->getUntitledFilename();
    SetFilename (untitledFilename, true);
  } else {
    if (! m_projectionFile.read (filename.c_str())) {
      *theApp->getLog() << "Unable to read projection file " << filename << "\n";
      return false;
    }
    *theApp->getLog() << "Read projection file " << filename << "\n";
    SetFilename(filename, true);
  }
  Modify(false);
  UpdateAllViews();
  return true;
}

bool ProjectionFileDocument::IsModified(void) const
{
  return wxDocument::IsModified();
}

void ProjectionFileDocument::Modify(bool mod)
{
    wxDocument::Modify(mod);
}


bool ProjectionFileDocument::OnCloseDocument ()
{
  bool bReturn = wxDocument::OnCloseDocument();
  return bReturn;
}

// PhantomDocument

IMPLEMENT_DYNAMIC_CLASS(PhantomDocument, wxDocument)

bool PhantomDocument::OnOpenDocument(const wxString& filename)
{
  wxString myFilename = filename;
  if (wxFile::Exists (myFilename)) {
    m_phantom.createFromFile (myFilename);
    *theApp->getLog() << "Read phantom file " << filename << "\n";
  } else {
    myFilename.Replace (".phm", "");
    m_phantom.createFromPhantom (myFilename);
  }
  m_namePhantom = myFilename;
  SetFilename (myFilename, true);
  if (m_phantom.fail()) {
    *theApp->getLog() << "Failure creating phantom " << myFilename << "\n";
    return false;
  }
  m_idPhantom = m_phantom.id();
  Modify(false);
  UpdateAllViews();
  return true;
}

bool PhantomDocument::OnCloseDocument ()
{
  bool bReturn = wxDocument::OnCloseDocument();
  return bReturn;
}

bool PhantomDocument::IsModified(void) const
{
  return wxDocument::IsModified();
}

void PhantomDocument::Modify(bool mod)
{
    wxDocument::Modify(mod);
}


// PlotFileDocument

IMPLEMENT_DYNAMIC_CLASS(PlotFileDocument, wxDocument)

bool PlotFileDocument::OnSaveDocument(const wxString& filename)
{
  m_namePlot = filename.c_str();
  if (! m_plot.fileWrite (filename)) {
    *theApp->getLog() << "Unable to write plot file " << filename << "\n";
    return false;
  }
  *theApp->getLog() << "Wrote plot file " << filename << "\n";
  Modify(false);
  return true;
}

bool PlotFileDocument::OnOpenDocument(const wxString& filename)
{
  if (filename == "untitled.plt") {
    wxString untitledFilename = theApp->getUntitledFilename();
    SetFilename (untitledFilename, true);
  } else {
    if (! m_plot.fileRead (filename.c_str())) {
      *theApp->getLog() << "Unable to read plot file " << filename << "\n";
      return false;
    }
    *theApp->getLog() << "Read plot file " << filename << "\n";
    SetFilename(filename, true);
    m_namePlot = filename.c_str();
  }
  Modify(false);
  UpdateAllViews();
  PlotFileView* ifView = dynamic_cast<PlotFileView*>(GetFirstView());
  if (ifView)
	  ifView->OnUpdate(ifView, NULL);

  return true;
}

bool PlotFileDocument::OnCloseDocument ()
{
  bool bReturn = wxDocument::OnCloseDocument();
  return bReturn;
}

bool PlotFileDocument::IsModified(void) const
{
  return wxDocument::IsModified();
}

void PlotFileDocument::Modify(bool mod)
{
    wxDocument::Modify(mod);
}

