/*****************************************************************************
** FILE IDENTIFICATION
**
**   Name:          view.cpp
**   Purpose:       View & Canvas routines for CTSim program
**   Programmer:    Kevin Rosenberg
**   Date Started:  July 2000
**
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: views.cpp,v 1.39 2000/12/27 03:16:02 kevin Exp $
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#if !wxUSE_DOC_VIEW_ARCHITECTURE
#error You must set wxUSE_DOC_VIEW_ARCHITECTURE to 1 in setup.h!
#endif

#include "wx/image.h"
#include "wx/progdlg.h"

#include "ct.h"
#include "ctsim.h"
#include "docs.h"
#include "views.h"
#include "dialogs.h"
#include "dlgprojections.h"
#include "dlgreconstruct.h"
#include "backprojectors.h"
#include "reconstruct.h"
#include "timer.h"

#if defined(MSVC) || HAVE_SSTREAM
#include <sstream>
#else
#include <sstream_subst>
#endif


// ImageFileCanvas

BEGIN_EVENT_TABLE(ImageFileCanvas, wxScrolledWindow)
EVT_MOUSE_EVENTS(ImageFileCanvas::OnMouseEvent)
END_EVENT_TABLE()


ImageFileCanvas::ImageFileCanvas (ImageFileView* v, wxFrame *frame, const wxPoint& pos, const wxSize& size, const long style)
: wxScrolledWindow(frame, -1, pos, size, style)
{
  m_pView = v;
  m_xCursor = -1;
  m_yCursor = -1;
}

void 
ImageFileCanvas::OnDraw(wxDC& dc)
{
  if (m_pView)
    m_pView->OnDraw(& dc);
}

void 
ImageFileCanvas::DrawRubberBandCursor (wxDC& dc, int x, int y)
{
  const ImageFile& rIF = m_pView->GetDocument()->getImageFile();
  ImageFileArrayConst v = rIF.getArray();
  int nx = rIF.nx();
  int ny = rIF.ny();
  
  int yPt = ny - y - 1;
  dc.SetLogicalFunction (wxINVERT);
  dc.SetPen (*wxGREEN_PEN);
  dc.DrawLine (0, yPt, nx, yPt);
  dc.DrawLine (x, 0, x, ny);
  dc.SetLogicalFunction (wxCOPY);
}

bool
ImageFileCanvas::GetCurrentCursor (int& x, int& y)
{
  x = m_xCursor;
  y = m_yCursor;
  
  if (m_xCursor >= 0 && m_yCursor >= 0)
    return true;
  else
    return false;
}

void 
ImageFileCanvas::OnMouseEvent(wxMouseEvent& event)
{
  if (! m_pView)
    return;
  
  wxClientDC dc(this);
  PrepareDC(dc);
  
  wxPoint pt(event.GetLogicalPosition(dc));
  
  const ImageFile& rIF = m_pView->GetDocument()->getImageFile();
  ImageFileArrayConst v = rIF.getArray();
  int nx = rIF.nx();
  int ny = rIF.ny();
  const int yPt = ny - 1 - pt.y;
  if (event.RightIsDown()) {
    if (pt.x >= 0 && pt.x < nx && pt.y >= 0 && pt.y < ny) {
      std::ostringstream os;
      os << "Image value (" << pt.x << "," << yPt << ") = " << v[pt.x][yPt] << "\n";
      *theApp->getLog() << os.str().c_str();
    } else
      *theApp->getLog() << "Mouse out of image range (" << pt.x << "," << yPt << ")\n";
  }
  else if (event.LeftIsDown() || event.LeftUp() || event.RightUp()) {
    if (pt.x >= 0 && pt.x < nx && pt.y >= 0 && pt.y < ny) {
      if (m_xCursor >= 0 && m_yCursor >= 0) {
        DrawRubberBandCursor (dc, m_xCursor, m_yCursor);
      }
      DrawRubberBandCursor (dc, pt.x, yPt);
      m_xCursor = pt.x;
      m_yCursor = yPt;
    } else
      *theApp->getLog() << "Mouse out of image range (" << pt.x << "," << yPt << ")\n";
  }
  if (event.LeftUp()) {
    std::ostringstream os;
    os << "Selected column " << pt.x << " , row " << yPt << "\n";
    *theApp->getLog() << os.str().c_str();
  }
}

// ImageFileView

IMPLEMENT_DYNAMIC_CLASS(ImageFileView, wxView)

BEGIN_EVENT_TABLE(ImageFileView, wxView)
EVT_MENU(IFMENU_FILE_PROPERTIES, ImageFileView::OnProperties)
EVT_MENU(IFMENU_VIEW_SCALE_MINMAX, ImageFileView::OnScaleMinMax)
EVT_MENU(IFMENU_VIEW_SCALE_AUTO, ImageFileView::OnScaleAuto)
EVT_MENU(IFMENU_COMPARE_IMAGES, ImageFileView::OnCompare)
EVT_MENU(IFMENU_COMPARE_ROW, ImageFileView::OnCompareRow)
EVT_MENU(IFMENU_COMPARE_COL, ImageFileView::OnCompareCol)
EVT_MENU(IFMENU_PLOT_ROW, ImageFileView::OnPlotRow)
EVT_MENU(IFMENU_PLOT_COL, ImageFileView::OnPlotCol)
END_EVENT_TABLE()

ImageFileView::ImageFileView(void) 
: wxView(), m_canvas(NULL), m_frame(NULL), m_bMinSpecified(false), m_bMaxSpecified(false)
{
}

ImageFileView::~ImageFileView(void)
{
}

void
ImageFileView::OnProperties (wxCommandEvent& event)
{
  double min, max, mean, mode, median, stddev;
  const ImageFile& rIF = GetDocument()->getImageFile();
  if (rIF.nx() == 0 || rIF.ny() == 0)
    *theApp->getLog() << "Properties: empty imagefile\n";
  else {
    const std::string& rFilename = rIF.getFilename();
    rIF.statistics (min, max, mean, mode, median, stddev);
    std::ostringstream os;
    os << "file: " << rFilename << "\nmin: "<<min<<"\nmax: "<<max<<"\nmean: "<<mean<<"\nmedian: "<<median<<"\nmode: "<<mode<<"\nstddev: "<<stddev << "\n";
    *theApp->getLog() << os.str().c_str();
    wxMessageDialog dialogMsg (m_frame, os.str().c_str(), "Imagefile Properties", wxOK | wxICON_INFORMATION);
    dialogMsg.ShowModal();
  }
}

void 
ImageFileView::OnScaleAuto (wxCommandEvent& event)
{
  const ImageFile& rIF = GetDocument()->getImageFile();
  double min, max, mean, mode, median, stddev;
  rIF.statistics(min, max, mean, mode, median, stddev);
  DialogAutoScaleParameters dialogAutoScale (m_frame, mean, mode, median, stddev, m_dAutoScaleFactor);
  int iRetVal = dialogAutoScale.ShowModal();
  if (iRetVal == wxID_OK) {
    m_bMinSpecified = true;
    m_bMaxSpecified = true;
    double dMin, dMax;
    if (dialogAutoScale.getMinMax (&dMin, &dMax)) {
      m_dMinPixel = dMin;
      m_dMaxPixel = dMax;
      m_dAutoScaleFactor = dialogAutoScale.getAutoScaleFactor();
      OnUpdate (this, NULL);
    }
  }
}

void 
ImageFileView::OnScaleMinMax (wxCommandEvent& event)
{
  const ImageFile& rIF = GetDocument()->getImageFile();
  double min, max;
  if (! m_bMinSpecified && ! m_bMaxSpecified)
    rIF.getMinMax (min, max);
  
  if (m_bMinSpecified)
    min = m_dMinPixel;
  if (m_bMaxSpecified)
    max = m_dMaxPixel;
  
  DialogGetMinMax dialogMinMax (m_frame, "Set Image Minimum & Maximum", min, max);
  int retVal = dialogMinMax.ShowModal();
  if (retVal == wxID_OK) {
    m_bMinSpecified = true;
    m_bMaxSpecified = true;
    m_dMinPixel = dialogMinMax.getMinimum();
    m_dMaxPixel = dialogMinMax.getMaximum();
    OnUpdate (this, NULL);
  }
}

void
ImageFileView::OnCompare (wxCommandEvent& event)
{
  std::vector<ImageFileDocument*> vecIF;
  theApp->getCompatibleImages (GetDocument(), vecIF);

  if (vecIF.size() == 0) {
    wxMessageBox("There are no compatible image files open for comparision", "No comparison images");
  } else {
    DialogGetComparisonImage dialogGetCompare(m_frame, "Get Comparison Image", vecIF, true);

    if (dialogGetCompare.ShowModal() == wxID_OK) {
      ImageFileDocument* pCompareDoc = dialogGetCompare.getImageFileDocument();
      const ImageFile& rIF = GetDocument()->getImageFile();
      const ImageFile& rCompareIF = pCompareDoc->getImageFile();
      std::ostringstream os;
      double min, max, mean, mode, median, stddev;
      rIF.statistics (min, max, mean, mode, median, stddev);
      os << rIF.getFilename() << ": minimum=" << min << ", maximum=" << max << ", mean=" << mean << ", mode=" << mode << ", median=" << median << ", stddev=" << stddev << "\n";
      rCompareIF.statistics (min, max, mean, mode, median, stddev);
      os << rCompareIF.getFilename() << ": minimum=" << min << ", maximum=" << max << ", mean=" << mean << ", mode=" << mode << ", median=" << median << ", stddev=" << stddev << "\n";
      os << "\n";
      double d, r, e;
      rIF.comparativeStatistics (rCompareIF, d, r, e);
      os << "Comparative Statistics: d=" << d << ", r=" << r << ", e=" << e << "\n";
      *theApp->getLog() << os.str().c_str();
      if (dialogGetCompare.getMakeDifferenceImage()) {
        ImageFileDocument* pDifferenceDoc = dynamic_cast<ImageFileDocument*>(theApp->getDocManager()->CreateDocument("untitled.if", wxDOC_SILENT));
        if (! pDifferenceDoc) {
          sys_error (ERR_SEVERE, "Unable to create image file");
          return;
        }
        ImageFile& differenceImage = pDifferenceDoc->getImageFile();
      
        differenceImage.setArraySize (rIF.nx(), rIF.ny());
        if (! rIF.subtractImages (rCompareIF, differenceImage)) {
          pDifferenceDoc->DeleteAllViews();
          return;
        }

        if (theApp->getSetModifyNewDocs())
          pDifferenceDoc->Modify(true);
        pDifferenceDoc->UpdateAllViews(this);
        pDifferenceDoc->GetFirstView()->OnUpdate (this, NULL);
      }
      wxMessageBox(os.str().c_str(), "Image Comparison");
    }
  }
}

ImageFileCanvas* 
ImageFileView::CreateCanvas (wxView *view, wxFrame *parent)
{
  ImageFileCanvas* pCanvas;
  int width, height;
  parent->GetClientSize(&width, &height);
  
  pCanvas = new ImageFileCanvas (dynamic_cast<ImageFileView*>(view), parent, wxPoint(0, 0), wxSize(width, height), 0);
  
  pCanvas->SetScrollbars(20, 20, 50, 50);
  pCanvas->SetBackgroundColour(*wxWHITE);
  pCanvas->Clear();
  
  return pCanvas;
}

wxFrame*
ImageFileView::CreateChildFrame(wxDocument *doc, wxView *view)
{
  wxDocChildFrame *subframe = new wxDocChildFrame(doc, view, theApp->getMainFrame(), -1, "ImageFile Frame", wxPoint(-1, -1), wxSize(0, 0), wxDEFAULT_FRAME_STYLE);
  
  wxMenu *file_menu = new wxMenu;
  
  file_menu->Append(MAINMENU_FILE_CREATE_PHANTOM, "Cr&eate Phantom...");
  file_menu->Append(wxID_OPEN, "&Open...");
  file_menu->Append(wxID_SAVE, "&Save");
  file_menu->Append(wxID_SAVEAS, "Save &As...");
  file_menu->Append(wxID_CLOSE, "&Close");
  
  file_menu->AppendSeparator();
  file_menu->Append(IFMENU_FILE_PROPERTIES, "P&roperties");
  
  file_menu->AppendSeparator();
  file_menu->Append(wxID_PRINT, "&Print...");
  file_menu->Append(wxID_PRINT_SETUP, "Print &Setup...");
  file_menu->Append(wxID_PREVIEW, "Print Pre&view");
  
  wxMenu *view_menu = new wxMenu;
  view_menu->Append(IFMENU_VIEW_SCALE_MINMAX, "Display Scale &Set...");
  view_menu->Append(IFMENU_VIEW_SCALE_AUTO, "Display Scale &Auto...");
  
  wxMenu *plot_menu = new wxMenu;
  plot_menu->Append (IFMENU_PLOT_ROW, "Plot &Row");
  plot_menu->Append (IFMENU_PLOT_COL, "Plot &Column");
  
  wxMenu *compare_menu = new wxMenu;
  compare_menu->Append (IFMENU_COMPARE_IMAGES, "Compare &Images...");
  compare_menu->Append (IFMENU_COMPARE_ROW, "Compare &Row");
  compare_menu->Append (IFMENU_COMPARE_COL, "Compare &Column");

  wxMenu *help_menu = new wxMenu;
  help_menu->Append(MAINMENU_HELP_ABOUT, "&About");
  
  wxMenuBar *menu_bar = new wxMenuBar;
  
  menu_bar->Append(file_menu, "&File");
  menu_bar->Append(view_menu, "&View");
  menu_bar->Append(plot_menu, "&Plot");
  menu_bar->Append(compare_menu, "&Compare");
  menu_bar->Append(help_menu, "&Help");
  
  subframe->SetMenuBar(menu_bar);
  
  subframe->Centre(wxBOTH);
  
  return subframe;
}


bool 
ImageFileView::OnCreate(wxDocument *doc, long WXUNUSED(flags) )
{
  m_frame = CreateChildFrame(doc, this);
  SetFrame (m_frame);
  
  m_bMinSpecified = false;
  m_bMaxSpecified = false;
  m_dAutoScaleFactor = 1.;
  
  int width, height;
  m_frame->GetClientSize (&width, &height);
  m_frame->SetTitle("ImageFileView");
  m_canvas = CreateCanvas (this, m_frame);
  
  int x, y;  // X requires a forced resize
  m_frame->GetSize(&x, &y);
  m_frame->SetSize(-1, -1, x, y);
  m_frame->SetFocus();
  m_frame->Show(true);
  Activate(true);
  
  return true;
}

void 
ImageFileView::OnDraw (wxDC* dc)
{
  if (m_bitmap.Ok())
    dc->DrawBitmap(m_bitmap, 0, 0, false);
  
  int xCursor, yCursor;
  if (m_canvas->GetCurrentCursor (xCursor, yCursor))
    m_canvas->DrawRubberBandCursor (*dc, xCursor, yCursor);
}


void 
ImageFileView::OnUpdate (wxView *WXUNUSED(sender), wxObject *WXUNUSED(hint) )
{
  const ImageFile& rIF = dynamic_cast<ImageFileDocument*>(GetDocument())->getImageFile();
  ImageFileArrayConst v = rIF.getArray();
  int nx = rIF.nx();
  int ny = rIF.ny();
  if (v != NULL && nx != 0 && ny != 0) {
    if (! m_bMinSpecified || ! m_bMaxSpecified) {
      double min, max;
      rIF.getMinMax (min, max);
      if (! m_bMinSpecified)
        m_dMinPixel = min;
      if (! m_bMaxSpecified)
        m_dMaxPixel = max;
    }
    double scaleWidth = m_dMaxPixel - m_dMinPixel;
    
    unsigned char* imageData = new unsigned char [nx * ny * 3];
    for (int ix = 0; ix < nx; ix++) {
      for (int iy = 0; iy < ny; iy++) {
        double scaleValue = ((v[ix][iy] - m_dMinPixel) / scaleWidth) * 255;
        int intensity = static_cast<int>(scaleValue + 0.5);
        intensity = clamp (intensity, 0, 255);
        int baseAddr = ((ny - 1 - iy) * nx + ix) * 3;
        imageData[baseAddr] = imageData[baseAddr+1] = imageData[baseAddr+2] = intensity;
      }
    }
    wxImage image (nx, ny, imageData, true);
    m_bitmap = image.ConvertToBitmap();
    delete imageData;
    int xSize = nx;
    int ySize = ny;
    xSize = clamp (xSize, 0, 800);
    ySize = clamp (ySize, 0, 800);
    m_frame->SetClientSize (xSize, ySize);
    m_canvas->SetScrollbars(20, 20, nx/20, ny/20);
    m_canvas->SetBackgroundColour(*wxWHITE);
  } 
  
  if (m_canvas)
    m_canvas->Refresh();
}

bool 
ImageFileView::OnClose (bool deleteWindow)
{
  if (!GetDocument()->Close())
    return false;
  
  m_canvas->Clear();
  m_canvas->m_pView = NULL;
  m_canvas = NULL;
  wxString s(theApp->GetAppName());
  if (m_frame)
    m_frame->SetTitle(s);
  SetFrame(NULL);
  
  Activate(false);
  
  if (deleteWindow) {
    delete m_frame;
    return true;
  }
  return true;
}


void
ImageFileView::OnPlotRow (wxCommandEvent& event)
{
  int xCursor, yCursor;
  if (! m_canvas->GetCurrentCursor (xCursor, yCursor)) {
    wxMessageBox ("No row selected. Please use left mouse button on image to select column","Error");
    return;
  }
  
  const ImageFile& rIF = dynamic_cast<ImageFileDocument*>(GetDocument())->getImageFile();
  ImageFileArrayConst v = rIF.getArray();
  int nx = rIF.nx();
  int ny = rIF.ny();
  
  if (v != NULL && yCursor < ny) {
    double* pX = new double [nx];
    double* pY = new double [nx];
    for (int i = 0; i < nx; i++) {
      pX[i] = i;
      pY[i] = v[i][yCursor];
    }
    PlotFileDocument* pPlotDoc = dynamic_cast<PlotFileDocument*>(theApp->getDocManager()->CreateDocument("untitled.plt", wxDOC_SILENT));
    if (! pPlotDoc) {
      sys_error (ERR_SEVERE, "Internal error: unable to create Plot file");
    } else {
      PlotFile& rPlotFile = pPlotDoc->getPlotFile();
      std::ostringstream os;
      os << "Row " << yCursor;
      std::string title("title ");
      title += os.str();
      rPlotFile.addEzsetCommand (title.c_str());
      rPlotFile.addEzsetCommand ("xlabel Column");
      rPlotFile.addEzsetCommand ("ylabel Pixel Value");
      rPlotFile.addEzsetCommand ("lxfrac 0");
      rPlotFile.addEzsetCommand ("box");
      rPlotFile.addEzsetCommand ("grid");
      rPlotFile.setCurveSize (2, nx);
      rPlotFile.addColumn (0, pX);
      rPlotFile.addColumn (1, pY);
    }
    delete pX;
    delete pY;
    if (theApp->getSetModifyNewDocs())
       pPlotDoc->Modify(true);
  }
  
}

void
ImageFileView::OnPlotCol (wxCommandEvent& event)
{
  int xCursor, yCursor;
  if (! m_canvas->GetCurrentCursor (xCursor, yCursor)) {
    wxMessageBox ("No column selected. Please use left mouse button on image to select column","Error");
    return;
  }
  
  const ImageFile& rIF = dynamic_cast<ImageFileDocument*>(GetDocument())->getImageFile();
  ImageFileArrayConst v = rIF.getArray();
  int nx = rIF.nx();
  int ny = rIF.ny();
  
  if (v != NULL && xCursor < nx) {
    double* pX = new double [ny];
    double* pY = new double [ny];
    for (int i = 0; i < ny; i++) {
      pX[i] = i;
      pY[i] = v[xCursor][i];
    }
    PlotFileDocument* pPlotDoc = dynamic_cast<PlotFileDocument*>(theApp->getDocManager()->CreateDocument("untitled.plt", wxDOC_SILENT));
    if (! pPlotDoc) {
      sys_error (ERR_SEVERE, "Internal error: unable to create Plot file");
    } else {
      PlotFile& rPlotFile = pPlotDoc->getPlotFile();
      std::ostringstream os;
      os << "Column " << xCursor;
      std::string title("title ");
      title += os.str();
      rPlotFile.addEzsetCommand (title.c_str());
      rPlotFile.addEzsetCommand ("xlabel Row");
      rPlotFile.addEzsetCommand ("ylabel Pixel Value");
      rPlotFile.addEzsetCommand ("lxfrac 0");
      rPlotFile.addEzsetCommand ("box");
      rPlotFile.addEzsetCommand ("grid");
      rPlotFile.setCurveSize (2, nx);
      rPlotFile.addColumn (0, pX);
      rPlotFile.addColumn (1, pY);
    }
    delete pX;
    delete pY;
    if (theApp->getSetModifyNewDocs())
      pPlotDoc->Modify(true);
  }
}

void
ImageFileView::OnCompareCol (wxCommandEvent& event)
{
  int xCursor, yCursor;
  if (! m_canvas->GetCurrentCursor (xCursor, yCursor)) {
    wxMessageBox ("No column selected. Please use left mouse button on image to select column","Error");
    return;
  }
  
  std::vector<ImageFileDocument*> vecIFDoc;
  theApp->getCompatibleImages (GetDocument(), vecIFDoc);
  if (vecIFDoc.size() == 0) {
    wxMessageBox ("No compatible images for Column Comparison", "Error");
    return;
  }
  DialogGetComparisonImage dialogGetCompare (m_frame, "Get Comparison Image", vecIFDoc, false);
  
  if (dialogGetCompare.ShowModal() == wxID_OK) {
    ImageFileDocument* pCompareDoc = dialogGetCompare.getImageFileDocument();
    const ImageFile& rIF = GetDocument()->getImageFile();
    const ImageFile& rCompareIF = pCompareDoc->getImageFile();
    
    ImageFileArrayConst v1 = rIF.getArray();
    ImageFileArrayConst v2 = rCompareIF.getArray();
    int nx = rIF.nx();
    int ny = rIF.ny();
    
    if (v1 != NULL && xCursor < nx) {
      double* pX = new double [ny];
      double* pY1 = new double [ny];
      double* pY2 = new double [ny];
      for (int i = 0; i < ny; i++) {
        pX[i] = i;
        pY1[i] = v1[xCursor][i];
        pY2[i] = v2[xCursor][i];
      }
      PlotFileDocument* pPlotDoc = dynamic_cast<PlotFileDocument*>(theApp->getDocManager()->CreateDocument("untitled.plt", wxDOC_SILENT));
      if (! pPlotDoc) {
        sys_error (ERR_SEVERE, "Internal error: unable to create Plot file");
      } else {
        PlotFile& rPlotFile = pPlotDoc->getPlotFile();
        std::ostringstream os;
        os << "Column " << xCursor << " Comparison";
        std::string title("title ");
        title += os.str();
        rPlotFile.addEzsetCommand (title.c_str());
        rPlotFile.addEzsetCommand ("xlabel Row");
        rPlotFile.addEzsetCommand ("ylabel Pixel Value");
        rPlotFile.addEzsetCommand ("lxfrac 0");
        rPlotFile.addEzsetCommand ("curve 1");
        rPlotFile.addEzsetCommand ("color 2");
        rPlotFile.addEzsetCommand ("curve 2");
        rPlotFile.addEzsetCommand ("color 4");
        rPlotFile.addEzsetCommand ("dash 5");
        rPlotFile.addEzsetCommand ("box");
        rPlotFile.addEzsetCommand ("grid");
        rPlotFile.setCurveSize (3, nx);
        rPlotFile.addColumn (0, pX);
        rPlotFile.addColumn (1, pY1);
        rPlotFile.addColumn (2, pY2);
      }
      delete pX;
      delete pY1;
      delete pY2;
      if (theApp->getSetModifyNewDocs())
        pPlotDoc->Modify(true);
    }
  }
}

void
ImageFileView::OnCompareRow (wxCommandEvent& event)
{
  int xCursor, yCursor;
  if (! m_canvas->GetCurrentCursor (xCursor, yCursor)) {
    wxMessageBox ("No column selected. Please use left mouse button on image to select column","Error");
    return;
  }
  
  std::vector<ImageFileDocument*> vecIFDoc;
  theApp->getCompatibleImages (GetDocument(), vecIFDoc);
  
  if (vecIFDoc.size() == 0) {
    wxMessageBox ("No compatible images for Row Comparison", "Error");
    return;
  }

  DialogGetComparisonImage dialogGetCompare (m_frame, "Get Comparison Image", vecIFDoc, false);
  
  if (dialogGetCompare.ShowModal() == wxID_OK) {
    ImageFileDocument* pCompareDoc = dialogGetCompare.getImageFileDocument();
    const ImageFile& rIF = GetDocument()->getImageFile();
    const ImageFile& rCompareIF = pCompareDoc->getImageFile();
    
    ImageFileArrayConst v1 = rIF.getArray();
    ImageFileArrayConst v2 = rCompareIF.getArray();
    int nx = rIF.nx();
    int ny = rIF.ny();
    
    if (v1 != NULL && yCursor < ny) {
      double* pX = new double [nx];
      double* pY1 = new double [nx];
      double* pY2 = new double [nx];
      for (int i = 0; i < nx; i++) {
        pX[i] = i;
        pY1[i] = v1[i][yCursor];
        pY2[i] = v2[i][yCursor];
      }
      PlotFileDocument* pPlotDoc = dynamic_cast<PlotFileDocument*>(theApp->getDocManager()->CreateDocument("untitled.plt", wxDOC_SILENT));
      if (! pPlotDoc) {
        sys_error (ERR_SEVERE, "Internal error: unable to create Plot file");
      } else {
        PlotFile& rPlotFile = pPlotDoc->getPlotFile();
        std::ostringstream os;
        os << "Row " << yCursor << " Comparison";
        std::string title("title ");
        title += os.str();
        rPlotFile.addEzsetCommand (title.c_str());
        rPlotFile.addEzsetCommand ("xlabel Column");
        rPlotFile.addEzsetCommand ("ylabel Pixel Value");
        rPlotFile.addEzsetCommand ("lxfrac 0");
        rPlotFile.addEzsetCommand ("curve 1");
        rPlotFile.addEzsetCommand ("color 2");
        rPlotFile.addEzsetCommand ("curve 2");
        rPlotFile.addEzsetCommand ("color 4");
        rPlotFile.addEzsetCommand ("dash 5");
        rPlotFile.addEzsetCommand ("box");
        rPlotFile.addEzsetCommand ("grid");
        rPlotFile.setCurveSize (3, ny);
        rPlotFile.addColumn (0, pX);
        rPlotFile.addColumn (1, pY1);
        rPlotFile.addColumn (2, pY2);
      }
      delete pX;
      delete pY1;
      delete pY2;
      if (theApp->getSetModifyNewDocs())
        pPlotDoc->Modify(true);
    }
  }
}


// PhantomCanvas

PhantomCanvas::PhantomCanvas (PhantomView* v, wxFrame *frame, const wxPoint& pos, const wxSize& size, const long style)
: wxScrolledWindow(frame, -1, pos, size, style)
{
  m_pView = v;
}

void 
PhantomCanvas::OnDraw(wxDC& dc)
{
  if (m_pView)
    m_pView->OnDraw(& dc);
}


// PhantomView

IMPLEMENT_DYNAMIC_CLASS(PhantomView, wxView)

BEGIN_EVENT_TABLE(PhantomView, wxView)
EVT_MENU(PHMMENU_FILE_PROPERTIES, PhantomView::OnProperties)
EVT_MENU(PHMMENU_PROCESS_RASTERIZE, PhantomView::OnRasterize)
EVT_MENU(PHMMENU_PROCESS_PROJECTIONS, PhantomView::OnProjections)
END_EVENT_TABLE()

PhantomView::PhantomView(void) 
: wxView(), m_canvas(NULL), m_frame(NULL)
{
  m_iDefaultNDet = 367;
  m_iDefaultNView = 320;
  m_iDefaultNSample = 2;
  m_dDefaultRotation = 2;
  m_dDefaultFocalLength = 2;
  m_dDefaultFieldOfView = 1;
  m_iDefaultGeometry = Scanner::GEOMETRY_PARALLEL;
  m_iDefaultTrace = Trace::TRACE_NONE;
}

PhantomView::~PhantomView(void)
{
}

void
PhantomView::OnProperties (wxCommandEvent& event)
{
  const int idPhantom = GetDocument()->getPhantomID();
  const wxString& namePhantom = GetDocument()->getPhantomName();
  std::ostringstream os;
  os << "Phantom " << namePhantom.c_str() << " (" << idPhantom << ")\n";
  *theApp->getLog() << os.str().c_str();
#if DEBUG
  const Phantom& rPhantom = GetDocument()->getPhantom();
  rPhantom.print();
#endif
}


void
PhantomView::OnProjections (wxCommandEvent& event)
{
  DialogGetProjectionParameters dialogProjection (m_frame, m_iDefaultNDet, m_iDefaultNView, m_iDefaultNSample, m_dDefaultRotation, m_dDefaultFocalLength, m_dDefaultFieldOfView, m_iDefaultGeometry, m_iDefaultTrace);
  int retVal = dialogProjection.ShowModal();
  if (retVal == wxID_OK) {
    m_iDefaultNDet = dialogProjection.getNDet();
    m_iDefaultNView = dialogProjection.getNView();
    m_iDefaultNSample = dialogProjection.getNSamples();
    m_iDefaultTrace = dialogProjection.getTrace();
    m_dDefaultRotation = dialogProjection.getRotAngle();
    m_dDefaultFocalLength = dialogProjection.getFocalLengthRatio();
    m_dDefaultFieldOfView = dialogProjection.getFieldOfViewRatio();
    wxString sGeometry = dialogProjection.getGeometry();
    m_iDefaultGeometry = Scanner::convertGeometryNameToID (sGeometry.c_str());
    
    if (m_iDefaultNDet > 0 && m_iDefaultNView > 0 && sGeometry != "") {
      const Phantom& rPhantom = GetDocument()->getPhantom();
      ProjectionFileDocument* pProjectionDoc = dynamic_cast<ProjectionFileDocument*>(theApp->getDocManager()->CreateDocument("untitled.pj", wxDOC_SILENT));
      if (! pProjectionDoc) {
        sys_error (ERR_SEVERE, "Unable to create projection document");
        return;
      }
      Projections& rProj = pProjectionDoc->getProjections();
      Scanner theScanner (rPhantom, sGeometry.c_str(), m_iDefaultNDet, m_iDefaultNView, m_iDefaultNSample, m_dDefaultRotation, m_dDefaultFocalLength, m_dDefaultFieldOfView);
      if (theScanner.fail()) {
        *theApp->getLog() << "Failed making scanner: " << theScanner.failMessage().c_str() << "\n";
        return;
      }
      rProj.initFromScanner (theScanner);
      m_dDefaultRotation /= PI;  // convert back to PI units
      
      if (m_iDefaultTrace > Trace::TRACE_CONSOLE) {
        ProjectionsDialog dialogProjections (theScanner, rProj, rPhantom, m_iDefaultTrace, dynamic_cast<wxWindow*>(m_frame));
        for (int iView = 0; iView < rProj.nView(); iView++) {
          ::wxYield();
          ::wxYield();
          if (dialogProjections.isCancelled() || ! dialogProjections.projectView (iView)) {
            pProjectionDoc->DeleteAllViews();
            return;
          }
          ::wxYield();
          ::wxYield();
          while (dialogProjections.isPaused()) {
            ::wxYield();
            ::wxUsleep(50);
          }
        }
      } else {
        wxProgressDialog dlgProgress (wxString("Projection"), wxString("Projection Progress"), rProj.nView() + 1, m_frame, wxPD_CAN_ABORT);
        for (int i = 0; i < rProj.nView(); i++) {
          theScanner.collectProjections (rProj, rPhantom, i, 1, true, m_iDefaultTrace);
          if (! dlgProgress.Update (i+1)) {
            pProjectionDoc->DeleteAllViews();
            return;
          }
        }
      }
      
      std::ostringstream os;
      os << "Projections for " << rPhantom.name() << ": nDet=" << m_iDefaultNDet << ", nView=" << m_iDefaultNView << ", nSamples=" << m_iDefaultNSample << ", RotAngle=" << m_dDefaultRotation << ", FocalLengthRatio=" << m_dDefaultFocalLength << ", FieldOfViewRatio=" << m_dDefaultFieldOfView << ", Geometry=" << sGeometry.c_str();
      rProj.setRemark (os.str());
      *theApp->getLog() << os.str().c_str() << "\n";
      
      m_frame->Lower();
      ::wxYield();
      ProjectionFileView* projView = dynamic_cast<ProjectionFileView*>(pProjectionDoc->GetFirstView());
      if (projView) {
        projView->getFrame()->SetFocus();
        projView->OnUpdate (projView, NULL);
      }
      if (wxView* pView = pProjectionDoc->GetFirstView()) {
        if (wxFrame* pFrame = pView->GetFrame()) {
          pFrame->SetFocus();
          pFrame->Raise();
        }
        theApp->getDocManager()->ActivateView (pView, true, false);
      }
      ::wxYield();
      if (theApp->getSetModifyNewDocs())
        pProjectionDoc->Modify(true);
      pProjectionDoc->UpdateAllViews(this);
    }
  }
}


void
PhantomView::OnRasterize (wxCommandEvent& event)
{
  DialogGetRasterParameters dialogRaster (m_frame, 256, 256, 1);
  int retVal = dialogRaster.ShowModal();
  if (retVal == wxID_OK) {
    int xSize = dialogRaster.getXSize();
    int ySize = dialogRaster.getYSize();
    int nSamples = dialogRaster.getNSamples();
    if (nSamples < 1)
      nSamples = 1;
    if (xSize > 0 && ySize > 0) {
      const Phantom& rPhantom = GetDocument()->getPhantom();
      ImageFileDocument* pRasterDoc = dynamic_cast<ImageFileDocument*>(theApp->getDocManager()->CreateDocument("untitled.if", wxDOC_SILENT));
      if (! pRasterDoc) {
        sys_error (ERR_SEVERE, "Unable to create image file");
        return;
      }
      ImageFile& imageFile = pRasterDoc->getImageFile();
      
      imageFile.setArraySize (xSize, ySize);
      wxProgressDialog dlgProgress (wxString("Rasterize"), wxString("Rasterization Progress"), imageFile.nx() + 1, m_frame, wxPD_CAN_ABORT);
      for (unsigned int i = 0; i < imageFile.nx(); i++) {
        rPhantom.convertToImagefile (imageFile, nSamples, Trace::TRACE_NONE, i, 1, true);
        if (! dlgProgress.Update(i+1)) {
          pRasterDoc->DeleteAllViews();
          return;
        }
      }
      if (theApp->getSetModifyNewDocs())
        pRasterDoc->Modify(true);
      pRasterDoc->UpdateAllViews(this);
      ImageFileView* rasterView = dynamic_cast<ImageFileView*>(pRasterDoc->GetFirstView());
      if (rasterView) {
        rasterView->getFrame()->SetFocus();
        rasterView->OnUpdate (rasterView, NULL);
      }
      
      std::ostringstream os;
      os << "Rasterize Phantom " << rPhantom.name() << ": XSize=" << xSize << ", YSize=" << ySize << ", nSamples=" << nSamples << "\n";
      *theApp->getLog() << os.str().c_str();
    }
  }
}


PhantomCanvas* 
PhantomView::CreateCanvas (wxView *view, wxFrame *parent)
{
  PhantomCanvas* pCanvas;
  int width, height;
  parent->GetClientSize(&width, &height);
  
  pCanvas = new PhantomCanvas (dynamic_cast<PhantomView*>(view), parent, wxPoint(0, 0), wxSize(width, height), 0);
  
  pCanvas->SetBackgroundColour(*wxWHITE);
  pCanvas->Clear();
  
  return pCanvas;
}

wxFrame*
PhantomView::CreateChildFrame(wxDocument *doc, wxView *view)
{
  wxDocChildFrame *subframe = new wxDocChildFrame(doc, view, theApp->getMainFrame(), -1, "Phantom Frame", wxPoint(10, 10), wxSize(256, 256), wxDEFAULT_FRAME_STYLE);
  
  wxMenu *file_menu = new wxMenu;
  
  file_menu->Append(MAINMENU_FILE_CREATE_PHANTOM, "Cr&eate Phantom...");
  file_menu->Append(wxID_OPEN, "&Open...");
  file_menu->Append(wxID_CLOSE, "&Close");
  
  file_menu->AppendSeparator();
  file_menu->Append(PHMMENU_FILE_PROPERTIES, "P&roperties");
  
  file_menu->AppendSeparator();
  file_menu->Append(wxID_PRINT, "&Print...");
  file_menu->Append(wxID_PRINT_SETUP, "Print &Setup...");
  file_menu->Append(wxID_PREVIEW, "Print Pre&view");
  
  wxMenu *process_menu = new wxMenu;
  process_menu->Append(PHMMENU_PROCESS_RASTERIZE, "&Rasterize...");
  process_menu->Append(PHMMENU_PROCESS_PROJECTIONS, "&Projections...");
  
  wxMenu *help_menu = new wxMenu;
  help_menu->Append(MAINMENU_HELP_CONTENTS, "&Contents");
  help_menu->Append(MAINMENU_HELP_ABOUT, "&About");
  
  wxMenuBar *menu_bar = new wxMenuBar;
  
  menu_bar->Append(file_menu, "&File");
  menu_bar->Append(process_menu, "&Process");
  menu_bar->Append(help_menu, "&Help");
  
  subframe->SetMenuBar(menu_bar);
  
  subframe->Centre(wxBOTH);
  
  return subframe;
}


bool 
PhantomView::OnCreate(wxDocument *doc, long WXUNUSED(flags) )
{
  m_frame = CreateChildFrame(doc, this);
  SetFrame(m_frame);
  
  int width, height;
  m_frame->GetClientSize(&width, &height);
  m_frame->SetTitle("PhantomView");
  m_canvas = CreateCanvas(this, m_frame);
  
#ifdef __X__
  int x, y;  // X requires a forced resize
  m_frame->GetSize(&x, &y);
  m_frame->SetSize(-1, -1, x, y);
#endif
  
  m_frame->Show(true);
  Activate(true);
  
  return true;
}


void 
PhantomView::OnUpdate(wxView *WXUNUSED(sender), wxObject *WXUNUSED(hint) )
{
  if (m_canvas)
    m_canvas->Refresh();
}

bool 
PhantomView::OnClose (bool deleteWindow)
{
  if (!GetDocument()->Close())
    return false;
  
  m_canvas->Clear();
  m_canvas->m_pView = NULL;
  m_canvas = NULL;
  wxString s(wxTheApp->GetAppName());
  if (m_frame)
    m_frame->SetTitle(s);
  SetFrame(NULL);
  
  Activate(false);
  
  if (deleteWindow) {
    delete m_frame;
    return true;
  }
  return true;
}

void
PhantomView::OnDraw (wxDC* dc)
{
  int xsize, ysize;
  m_canvas->GetClientSize (&xsize, &ysize);
  SGPDriver driver (dc, xsize, ysize);
  SGP sgp (driver);
  const Phantom& rPhantom = GetDocument()->getPhantom();
  sgp.setColor (C_RED);
  rPhantom.show (sgp);
}

// ProjectionCanvas

ProjectionFileCanvas::ProjectionFileCanvas (ProjectionFileView* v, wxFrame *frame, const wxPoint& pos, const wxSize& size, const long style)
: wxScrolledWindow(frame, -1, pos, size, style)
{
  m_pView = v;
}

void 
ProjectionFileCanvas::OnDraw(wxDC& dc)
{
  if (m_pView)
    m_pView->OnDraw(& dc);
}

// ProjectionFileView

IMPLEMENT_DYNAMIC_CLASS(ProjectionFileView, wxView)

BEGIN_EVENT_TABLE(ProjectionFileView, wxView)
EVT_MENU(PJMENU_FILE_PROPERTIES, ProjectionFileView::OnProperties)
EVT_MENU(PJMENU_PROCESS_RECONSTRUCT, ProjectionFileView::OnReconstruct)
END_EVENT_TABLE()

ProjectionFileView::ProjectionFileView(void) 
: wxView(), m_canvas(NULL), m_frame(NULL)
{
  m_iDefaultNX = 256;
  m_iDefaultNY = 256;
  m_iDefaultFilter = SignalFilter::FILTER_ABS_BANDLIMIT;
  m_dDefaultFilterParam = 1.;
#if HAVE_FFTW
  m_iDefaultFilterMethod = ProcessSignal::FILTER_METHOD_RFFTW;
  m_iDefaultFilterGeneration = ProcessSignal::FILTER_GENERATION_INVERSE_FOURIER;
#else
  m_iDefaultFilterMethod = ProcessSignal::FILTER_METHOD_CONVOLUTION;
  m_iDefaultFilterGeneration = ProcessSignal::FILTER_GENERATION_DIRECT;
#endif
  m_iDefaultZeropad = 1;
  m_iDefaultBackprojector = Backprojector::BPROJ_IDIFF3;
  m_iDefaultInterpolation = Backprojector::INTERP_LINEAR;
  m_iDefaultInterpParam = 1;
  m_iDefaultTrace = Trace::TRACE_NONE;
}

ProjectionFileView::~ProjectionFileView(void)
{
}

void
ProjectionFileView::OnProperties (wxCommandEvent& event)
{
  const Projections& rProj = GetDocument()->getProjections();
  std::ostringstream os;
  rProj.printScanInfo(os);
  *theApp->getLog() << os.str().c_str();
  wxMessageDialog dialogMsg (m_frame, os.str().c_str(), "Projection File Properties", wxOK | wxICON_INFORMATION);
  dialogMsg.ShowModal();
}


void
ProjectionFileView::OnReconstruct (wxCommandEvent& event)
{
  DialogGetReconstructionParameters dialogReconstruction (m_frame, m_iDefaultNX, m_iDefaultNY, m_iDefaultFilter, m_dDefaultFilterParam, m_iDefaultFilterMethod, m_iDefaultFilterGeneration, m_iDefaultZeropad, m_iDefaultInterpolation, m_iDefaultInterpParam, m_iDefaultBackprojector, m_iDefaultTrace);
  
  int retVal = dialogReconstruction.ShowModal();
  if (retVal == wxID_OK) {
    m_iDefaultNX = dialogReconstruction.getXSize();
    m_iDefaultNY = dialogReconstruction.getYSize();
    wxString optFilterName = dialogReconstruction.getFilterName();
    m_iDefaultFilter = SignalFilter::convertFilterNameToID (optFilterName.c_str());
    m_dDefaultFilterParam = dialogReconstruction.getFilterParam();
    wxString optFilterMethodName = dialogReconstruction.getFilterMethodName();
    m_iDefaultFilterMethod = ProcessSignal::convertFilterMethodNameToID(optFilterMethodName.c_str());
    m_iDefaultZeropad = dialogReconstruction.getZeropad();
    wxString optFilterGenerationName = dialogReconstruction.getFilterGenerationName();
    m_iDefaultFilterGeneration = ProcessSignal::convertFilterGenerationNameToID (optFilterGenerationName.c_str());
    wxString optInterpName = dialogReconstruction.getInterpName();
    m_iDefaultInterpolation = Backprojector::convertInterpNameToID (optInterpName.c_str());
    m_iDefaultInterpParam = dialogReconstruction.getInterpParam();
    wxString optBackprojectName = dialogReconstruction.getBackprojectName();
    m_iDefaultBackprojector = Backprojector::convertBackprojectNameToID (optBackprojectName.c_str());
    m_iDefaultTrace = dialogReconstruction.getTrace();
    if (m_iDefaultNX > 0 && m_iDefaultNY > 0) {
      ImageFileDocument* pReconDoc = dynamic_cast<ImageFileDocument*>(theApp->getDocManager()->CreateDocument("untitled.if", wxDOC_SILENT));
      if (! pReconDoc) {
        sys_error (ERR_SEVERE, "Unable to create image file");
        return;
      }
      ImageFile& imageFile = pReconDoc->getImageFile();
      const Projections& rProj = GetDocument()->getProjections();
      imageFile.setArraySize (m_iDefaultNX, m_iDefaultNY);
      
      if (m_iDefaultFilterMethod != ProcessSignal::FILTER_METHOD_CONVOLUTION && m_iDefaultFilterGeneration == ProcessSignal::FILTER_GENERATION_DIRECT && rProj.geometry() != Scanner::GEOMETRY_PARALLEL) {
        wxMessageBox ("Sorry!\nCurrently, frequency-based filtering with direct filter generation is not support for geometries other than parallel.\nAborting command.", "Not Supported", wxOK | wxICON_WARNING, m_frame);
        return;
      }
#if 0
      SGPDriver* pSGPDriver = NULL;
      SGP* pSGP = NULL;
      wxMemoryDC* pDCPlot = NULL;
      wxBitmap bitmap;
      if (m_iDefaultTrace >= Trace::TRACE_PLOT) {
        bitmap.Create (500, 500);
        pDCPlot = new wxMemoryDC;
        pDCPlot->SelectObject (bitmap);
        pSGPDriver = new SGPDriver (dynamic_cast<wxDC*>pDCPlot, 500, 500);
        pSGP = new SGP (*pSGPDriver);
      }
      Reconstructor* pReconstruct = new Reconstructor (rProj, imageFile, optFilterName.c_str(), m_dDefaultFilterParam, optFilterMethodName.c_str(), m_iDefaultZeropad, optFilterGenerationName.c_str(), optInterpName.c_str(), m_iDefaultInterpParam, optBackprojectName.c_str(), m_iDefaultTrace, pSGP);
      delete pSGP;
#else
      Reconstructor* pReconstruct = new Reconstructor (rProj, imageFile, optFilterName.c_str(), m_dDefaultFilterParam, optFilterMethodName.c_str(), m_iDefaultZeropad, optFilterGenerationName.c_str(), optInterpName.c_str(), m_iDefaultInterpParam, optBackprojectName.c_str(), m_iDefaultTrace);
#endif
      
      Timer timerRecon;
      if (m_iDefaultTrace > Trace::TRACE_CONSOLE) {
        ReconstructDialog* pDlgReconstruct = new ReconstructDialog (*pReconstruct, rProj, imageFile, m_iDefaultTrace, m_frame);
        for (int iView = 0; iView < rProj.nView(); iView++) {
          ::wxYield();
          ::wxYield();
          if (pDlgReconstruct->isCancelled() || ! pDlgReconstruct->reconstructView (iView)) {
            delete pDlgReconstruct;
            delete pReconstruct;
            pReconDoc->DeleteAllViews();
            return;
          }
          ::wxYield();
          ::wxYield();
          while (pDlgReconstruct->isPaused()) {
            ::wxYield();
            ::wxUsleep(50);
          }
        }
        delete pDlgReconstruct;
      } else {
        wxProgressDialog dlgProgress (wxString("Reconstruction"), wxString("Reconstruction Progress"), rProj.nView() + 1, m_frame, wxPD_CAN_ABORT);
        for (int i = 0; i < rProj.nView(); i++) {
          pReconstruct->reconstructView (i, 1);
          if (! dlgProgress.Update(i + 1)) {
            delete pReconstruct;
            pReconDoc->DeleteAllViews();
            return;
          }
        }
      }
      delete pReconstruct;
      if (theApp->getSetModifyNewDocs())
        pReconDoc->Modify(true);
      pReconDoc->UpdateAllViews(this);
      ImageFileView* rasterView = dynamic_cast<ImageFileView*>(pReconDoc->GetFirstView());
      if (rasterView) {
        rasterView->getFrame()->SetFocus();
        rasterView->OnUpdate (rasterView, NULL);
      }
      std::ostringstream os;
      os << "Reconstruct " << rProj.getFilename() << ": xSize=" << m_iDefaultNX << ", ySize=" << m_iDefaultNY << ", Filter=" << optFilterName.c_str() << ", FilterParam=" << m_dDefaultFilterParam << ", FilterMethod=" << optFilterMethodName.c_str() << ", FilterGeneration=" << optFilterGenerationName.c_str() << ", Zeropad=" << m_iDefaultZeropad << ", Interpolation=" << optInterpName.c_str() << ", InterpolationParam=" << m_iDefaultInterpParam << ", Backprojection=" << optBackprojectName.c_str();
      *theApp->getLog() << os.str().c_str() << "\n";
      imageFile.labelAdd (rProj.getLabel());
      imageFile.labelAdd (Array2dFileLabel::L_HISTORY, os.str().c_str(), timerRecon.timerEnd());
    }
  }
}


ProjectionFileCanvas* 
ProjectionFileView::CreateCanvas (wxView *view, wxFrame *parent)
{
  ProjectionFileCanvas* pCanvas;
  int width, height;
  parent->GetClientSize(&width, &height);
  
  pCanvas = new ProjectionFileCanvas (dynamic_cast<ProjectionFileView*>(view), parent, wxPoint(0, 0), wxSize(width, height), 0);
  
  pCanvas->SetScrollbars(20, 20, 50, 50);
  pCanvas->SetBackgroundColour(*wxWHITE);
  pCanvas->Clear();
  
  return pCanvas;
}

wxFrame*
ProjectionFileView::CreateChildFrame(wxDocument *doc, wxView *view)
{
  wxDocChildFrame *subframe = new wxDocChildFrame(doc, view, theApp->getMainFrame(), -1, "Projection Frame", wxPoint(10, 10), wxSize(0, 0), wxDEFAULT_FRAME_STYLE);
  
  wxMenu *file_menu = new wxMenu;
  
  file_menu->Append(MAINMENU_FILE_CREATE_PHANTOM, "Cr&eate Phantom...");
  file_menu->Append(wxID_OPEN, "&Open...");
  file_menu->Append(wxID_SAVE, "&Save");
  file_menu->Append(wxID_SAVEAS, "Save &As...");
  file_menu->Append(wxID_CLOSE, "&Close");
  
  file_menu->AppendSeparator();
  file_menu->Append(PJMENU_FILE_PROPERTIES, "P&roperties");
  
  file_menu->AppendSeparator();
  file_menu->Append(wxID_PRINT, "&Print...");
  file_menu->Append(wxID_PRINT_SETUP, "Print &Setup...");
  file_menu->Append(wxID_PREVIEW, "Print Pre&view");
  
  wxMenu *process_menu = new wxMenu;
  process_menu->Append(PJMENU_PROCESS_RECONSTRUCT, "R&econstruct...");
  
  wxMenu *help_menu = new wxMenu;
  help_menu->Append(MAINMENU_HELP_CONTENTS, "&Contents");
  help_menu->AppendSeparator();
  help_menu->Append(MAINMENU_HELP_ABOUT, "&About");
  
  wxMenuBar *menu_bar = new wxMenuBar;
  
  menu_bar->Append(file_menu, "&File");
  menu_bar->Append(process_menu, "&Process");
  menu_bar->Append(help_menu, "&Help");
  
  subframe->SetMenuBar(menu_bar);
  
  subframe->Centre(wxBOTH);
  
  return subframe;
}


bool 
ProjectionFileView::OnCreate(wxDocument *doc, long WXUNUSED(flags) )
{
  m_frame = CreateChildFrame(doc, this);
  SetFrame(m_frame);
  
  int width, height;
  m_frame->GetClientSize(&width, &height);
  m_frame->SetTitle("ProjectionFileView");
  m_canvas = CreateCanvas(this, m_frame);
  
#ifdef __X__
  int x, y;  // X requires a forced resize
  m_frame->GetSize(&x, &y);
  m_frame->SetSize(-1, -1, x, y);
#endif
  
  m_frame->Show(true);
  Activate(true);
  
  return true;
}

void 
ProjectionFileView::OnDraw (wxDC* dc)
{
  if (m_bitmap.Ok())
    dc->DrawBitmap (m_bitmap, 0, 0, false);
}


void 
ProjectionFileView::OnUpdate(wxView *WXUNUSED(sender), wxObject *WXUNUSED(hint) )
{
  const Projections& rProj = GetDocument()->getProjections();
  const int nDet = rProj.nDet();
  const int nView = rProj.nView();
  if (nDet != 0 && nView != 0) {
    const DetectorArray& detarray = rProj.getDetectorArray(0);
    const DetectorValue* detval = detarray.detValues();
    double min = detval[0];
    double max = detval[0];
    for (int iy = 0; iy < nView; iy++) {
      const DetectorArray& detarray = rProj.getDetectorArray(iy);
      const DetectorValue* detval = detarray.detValues();
      for (int ix = 0; ix < nDet; ix++) {
        if (min > detval[ix])
          min = detval[ix];
        else if (max < detval[ix])
          max = detval[ix];
      }
    }
    
    unsigned char* imageData = new unsigned char [nDet * nView * 3];
    double scale = (max - min) / 255;
    for (int iy2 = 0; iy2 < nView; iy2++) {
      const DetectorArray& detarray = rProj.getDetectorArray (iy2);
      const DetectorValue* detval = detarray.detValues();
      for (int ix = 0; ix < nDet; ix++) {
        int intensity = static_cast<int>(((detval[ix] - min) / scale) + 0.5);
        intensity = clamp(intensity, 0, 255);
        int baseAddr = (iy2 * nDet + ix) * 3;
        imageData[baseAddr] = imageData[baseAddr+1] = imageData[baseAddr+2] = intensity;
      }
    }
    wxImage image (nDet, nView, imageData, true);
    m_bitmap = image.ConvertToBitmap();
    delete imageData;
    int xSize = nDet;
    int ySize = nView;
    xSize = clamp (xSize, 0, 800);
    ySize = clamp (ySize, 0, 800);
    m_frame->SetClientSize (xSize, ySize);
    m_canvas->SetScrollbars (20, 20, nDet/20, nView/20);
  }
  
  if (m_canvas)
    m_canvas->Refresh();
}

bool 
ProjectionFileView::OnClose (bool deleteWindow)
{
  if (!GetDocument()->Close())
    return false;
  
  m_canvas->Clear();
  m_canvas->m_pView = NULL;
  m_canvas = NULL;
  wxString s(wxTheApp->GetAppName());
  if (m_frame)
    m_frame->SetTitle(s);
  SetFrame(NULL);
  
  Activate(false);
  
  if (deleteWindow) {
    delete m_frame;
    return true;
  }
  return true;
}



// PlotFileCanvas
PlotFileCanvas::PlotFileCanvas (PlotFileView* v, wxFrame *frame, const wxPoint& pos, const wxSize& size, const long style)
: wxScrolledWindow(frame, -1, pos, size, style)
{
  m_pView = v;
}

void 
PlotFileCanvas::OnDraw(wxDC& dc)
{
  if (m_pView)
    m_pView->OnDraw(& dc);
}


// PlotFileView

IMPLEMENT_DYNAMIC_CLASS(PlotFileView, wxView)

BEGIN_EVENT_TABLE(PlotFileView, wxView)
EVT_MENU(PJMENU_FILE_PROPERTIES, PlotFileView::OnProperties)
EVT_MENU(PLOTMENU_VIEW_SCALE_MINMAX, PlotFileView::OnScaleMinMax)
EVT_MENU(PLOTMENU_VIEW_SCALE_AUTO, PlotFileView::OnScaleAuto)
END_EVENT_TABLE()

PlotFileView::PlotFileView(void) 
: wxView(), m_canvas(NULL), m_frame(NULL)
{
}

PlotFileView::~PlotFileView(void)
{
}

void
PlotFileView::OnProperties (wxCommandEvent& event)
{
  const PlotFile& rProj = GetDocument()->getPlotFile();
  std::ostringstream os;
  os << "Columns: " << rProj.getNumColumns() << ", Records: " << rProj.getNumRecords() << "\n";
  *theApp->getLog() << os.str().c_str();
  wxMessageDialog dialogMsg (m_frame, os.str().c_str(), "Plot File Properties", wxOK | wxICON_INFORMATION);
  dialogMsg.ShowModal();
}


void 
PlotFileView::OnScaleAuto (wxCommandEvent& event)
{
  const PlotFile& rPlotFile = GetDocument()->getPlotFile();
  double min, max, mean, mode, median, stddev;
  rPlotFile.statistics (1, min, max, mean, mode, median, stddev);
  DialogAutoScaleParameters dialogAutoScale (m_frame, mean, mode, median, stddev, m_dAutoScaleFactor);
  int iRetVal = dialogAutoScale.ShowModal();
  if (iRetVal == wxID_OK) {
    m_bMinSpecified = true;
    m_bMaxSpecified = true;
    double dMin, dMax;
    if (dialogAutoScale.getMinMax (&dMin, &dMax)) {
      m_dMinPixel = dMin;
      m_dMaxPixel = dMax;
      m_dAutoScaleFactor = dialogAutoScale.getAutoScaleFactor();
      OnUpdate (this, NULL);
    }
  }
}

void 
PlotFileView::OnScaleMinMax (wxCommandEvent& event)
{
  const PlotFile& rPlotFile = GetDocument()->getPlotFile();
  double min, max;
  
  if (! m_bMinSpecified && ! m_bMaxSpecified) {
    if (! rPlotFile.getMinMax (1, min, max)) {
      *theApp->getLog() << "Error: unable to find Min/Max\n";
      return;
    }
  }
  
  if (m_bMinSpecified)
    min = m_dMinPixel;
  if (m_bMaxSpecified)
    max = m_dMaxPixel;
  
  DialogGetMinMax dialogMinMax (m_frame, "Set Y-axis Minimum & Maximum", min, max);
  int retVal = dialogMinMax.ShowModal();
  if (retVal == wxID_OK) {
    m_bMinSpecified = true;
    m_bMaxSpecified = true;
    m_dMinPixel = dialogMinMax.getMinimum();
    m_dMaxPixel = dialogMinMax.getMaximum();
    OnUpdate (this, NULL);
  }
}


PlotFileCanvas* 
PlotFileView::CreateCanvas (wxView *view, wxFrame *parent)
{
  PlotFileCanvas* pCanvas;
  int width, height;
  parent->GetClientSize(&width, &height);
  
  pCanvas = new PlotFileCanvas (dynamic_cast<PlotFileView*>(view), parent, wxPoint(0, 0), wxSize(width, height), 0);
  
  pCanvas->SetBackgroundColour(*wxWHITE);
  pCanvas->Clear();
  
  return pCanvas;
}

wxFrame*
PlotFileView::CreateChildFrame(wxDocument *doc, wxView *view)
{
  wxDocChildFrame *subframe = new wxDocChildFrame(doc, view, theApp->getMainFrame(), -1, "Plot Frame", wxPoint(10, 10), wxSize(500, 300), wxDEFAULT_FRAME_STYLE);
  
  wxMenu *file_menu = new wxMenu;
  
  file_menu->Append(MAINMENU_FILE_CREATE_PHANTOM, "Cr&eate Phantom...");
  file_menu->Append(wxID_OPEN, "&Open...");
  file_menu->Append(wxID_SAVE, "&Save");
  file_menu->Append(wxID_SAVEAS, "Save &As...");
  file_menu->Append(wxID_CLOSE, "&Close");
  
  file_menu->AppendSeparator();
  file_menu->Append(PJMENU_FILE_PROPERTIES, "P&roperties");
  
  file_menu->AppendSeparator();
  file_menu->Append(wxID_PRINT, "&Print...");
  file_menu->Append(wxID_PRINT_SETUP, "Print &Setup...");
  file_menu->Append(wxID_PREVIEW, "Print Pre&view");
  
  wxMenu *view_menu = new wxMenu;
  view_menu->Append(PLOTMENU_VIEW_SCALE_MINMAX, "Display Scale &Set...");
  view_menu->Append(PLOTMENU_VIEW_SCALE_AUTO, "Display Scale &Auto...");
  
  wxMenu *help_menu = new wxMenu;
  help_menu->Append(MAINMENU_HELP_CONTENTS, "&Contents");
  help_menu->AppendSeparator();
  help_menu->Append(MAINMENU_HELP_ABOUT, "&About");
  
  wxMenuBar *menu_bar = new wxMenuBar;
  
  menu_bar->Append(file_menu, "&File");
  menu_bar->Append(view_menu, "&View");
  menu_bar->Append(help_menu, "&Help");
  
  subframe->SetMenuBar(menu_bar);
  
  subframe->Centre(wxBOTH);
  
  return subframe;
}


bool 
PlotFileView::OnCreate(wxDocument *doc, long WXUNUSED(flags) )
{
  m_frame = CreateChildFrame(doc, this);
  SetFrame(m_frame);
  
  m_bMinSpecified = false;
  m_bMaxSpecified = false;
  m_dAutoScaleFactor = 1.;
  
  int width, height;
  m_frame->GetClientSize(&width, &height);
  m_frame->SetTitle ("Plot File");
  m_canvas = CreateCanvas (this, m_frame);
  
#ifdef __X__
  int x, y;  // X requires a forced resize
  m_frame->GetSize(&x, &y);
  m_frame->SetSize(-1, -1, x, y);
#endif
  
  m_frame->Show(true);
  Activate(true);
  
  return true;
}

void 
PlotFileView::OnDraw (wxDC* dc)
{
  const PlotFile& rPlotFile = GetDocument()->getPlotFile();
  const int iNColumns = rPlotFile.getNumColumns();
  const int iNRecords = rPlotFile.getNumRecords();
  
  if (iNColumns > 0 && iNRecords > 0) {
    int xsize, ysize;
    m_canvas->GetClientSize (&xsize, &ysize);
    SGPDriver driver (dc, xsize, ysize);
    SGP sgp (driver);
    EZPlot plot (sgp);

    for (int iEzset = 0; iEzset < rPlotFile.getNumEzsetCommands(); iEzset++)
       plot.ezset (rPlotFile.getEzsetCommand (iEzset));

    if (m_bMinSpecified) {
      std::ostringstream os;
      os << "ymin " << m_dMinPixel;
      plot.ezset (os.str());
    }
    
    if (m_bMaxSpecified) {
      std::ostringstream os;
      os << "ymax " << m_dMaxPixel;
      plot.ezset (os.str());
    }
    
    plot.ezset("box");
    plot.ezset("grid");
    
    double* pdXaxis = new double [iNRecords];
    rPlotFile.getColumn (0, pdXaxis);
    
    double* pdY = new double [iNRecords];
    for (int iCol = 1; iCol < iNColumns; iCol++) {
      rPlotFile.getColumn (iCol, pdY);
      plot.addCurve (pdXaxis, pdY, iNRecords);
    }
    
    delete pdXaxis;
    delete pdY;
    
    plot.plot();
  }
}


void 
PlotFileView::OnUpdate(wxView *WXUNUSED(sender), wxObject *WXUNUSED(hint) )
{
  if (m_canvas)
    m_canvas->Refresh();
}

bool 
PlotFileView::OnClose (bool deleteWindow)
{
  if (!GetDocument()->Close())
    return false;
  
  m_canvas->Clear();
  m_canvas->m_pView = NULL;
  m_canvas = NULL;
  wxString s(wxTheApp->GetAppName());
  if (m_frame)
    m_frame->SetTitle(s);
  SetFrame(NULL);
  
  Activate(false);
  
  if (deleteWindow) {
    delete m_frame;
    return true;
  }
  return true;
}

